#-----------------------------------------------------------------------------
#
#   Supersymmetry - World
#
#-----------------------------------------------------------------------------

import random, traceback
from math import ceil
from pygame import Rect
from albow.resource import get_sound
from albow.dialogs import alert, ask
from constants import *
from vectors import add, opposite_side
import tiles
import particles

beam_sound = get_sound("beam.ogg")
beam_sound.set_volume(0.4)

switch_on_sound = get_sound("switch-on.ogg")
switch_off_sound = get_sound("switch-off.ogg")
success_sound = get_sound("success.ogg")
skint_sound = get_sound("skint.ogg")
money_sound = get_sound("money.ogg")
game_over_sound = get_sound("applause.ogg")
champagne_sound = get_sound("champagne.ogg")
placement_sound = get_sound("place.ogg")
removal_sound = get_sound("remove.ogg")
rotation_sound = get_sound("rotate.ogg")

#-----------------------------------------------------------------------------

class World(object):
  #  level_number   int
  #  initial_grid   int[grid_size][grid_size]
	#  grid           int[grid_size][grid_size]
	#  start_cell     (col, row)
	#  initial_funds  int
	#  funds          int
	#  game_over      bool
	#  grid_bounds    Rect
	#  initial_inventory    [int]
	#  inventory            [int]
	#  initial_inventory_selection  int
	#  inventory_selection  int
	#  circuit_complete     bool
	#  vacuum_ok            bool
	#  beam_energy          int
	#  running              bool
	#  particles_found      int
	
	game_over = False
	initial_inventory = []
	vacuum_ok = False

	def __init__(self):
		self.level_number = 0
		self.funds = STARTING_FUNDS
		self.grid = make_grid()
		self.game_over = False
		self.inventory = [1, 3]
		self.inventory_selection = 0
		self.running = False
		self.particles_found = 0
		self.new_level()
		#self.dump_levels() ###
	
	def new_level(self):
		if self.particles_found >= particles.num_particles():
			self.game_over = True
			announce_success()
			return
		self.level_number += 1
		self.funds += self.funds_required_for_level(self.level_number)
		self.initial_funds = self.funds
		self.initial_grid = self.grid
		self.initial_inventory = self.inventory
		self.initial_inventory_selection = self.inventory_selection
		self.grid_bounds = grid_bounds_for_level(self.level_number)
		if self.level_number == 1:
			self.start_cell = (0, GRID_ROWS // 2)
			self.place_start_pipe()
		self.start_level()
	
	def place_start_pipe(self):
		col, row = self.start_cell
		self.initial_grid[row][col] = tiles.PIPE_START.tileno
	
	def start_level(self):
		self.grid = copy_grid(self.initial_grid)
		self.funds = self.initial_funds
		self.inventory = self.initial_inventory[:]
		self.inventory_selection = self.initial_inventory_selection
	
	def restart_level(self):
		if ask("Restart the current level?") == "OK":
			self.start_level()
	
	def get_tile(self, cell):
		x, y = cell
		if 0 <= x < GRID_COLS and 0 <= y < GRID_ROWS:
			return tiles.get(self.grid[y][x])
	
	def set_tile(self, cell, tile):
		x, y = cell
		if 0 <= x < GRID_COLS and 0 <= y < GRID_ROWS:
			self.grid[y][x] = tiles.tileno(tile)
			game.changed()
	
	def cell_in_bounds(self, cell):
		return self.grid_bounds.collidepoint(cell)
	
	def place_tile(self, cell):
		tile = self.get_tile(cell)
		if tile:
			if tile.fixed:
				return
			tile = tiles.rotate_tile(tile)
			self.set_tile(cell, tile)
			rotation_sound.play()
		else:
			if not self.have_funds(PLACEMENT_COST):
				return
			#print "TODO: World.place_tile: Play placement sound" ###
			tile = self.take_tile_from_inventory()
			if not tile:
				return
			self.spend_funds(PLACEMENT_COST)
			self.set_tile(cell, tile)
			placement_sound.play()
	
	def remove_tile(self, cell):
		tile = self.get_tile(cell)
		if tile and not tile.fixed and self.room_in_inventory():
			if self.spend_funds(PLACEMENT_COST):
				self.add_tile_to_inventory(tile)
				self.set_tile(cell, None)
				removal_sound.play()
	
	def room_in_inventory(self):
		return True
	
	def take_tile_from_inventory(self):
		inventory = self.inventory
		i = self.inventory_selection
		n = len(inventory)
		if 0 <= i < n:
			tileno = inventory.pop(i)
			n -= 1
			if i >= n:
				i = n - 1
			if n and inventory[i] <> tileno:
				if tileno in inventory:
					i = inventory.index(tileno)
			self.inventory_selection = i
			game.changed()
			return tiles.get(tileno)
	
	def add_tile_to_inventory(self, tile):
		if tile:
			inventory = self.inventory
			i = self.inventory_selection
			n = len(inventory)
			tileno = tile.tileno
			if tileno in inventory:
				j = inventory.index(tileno)
				while j < n and inventory[j] == tileno:
					j += 1
			else:
				j = len(inventory)
			inventory.insert(j, tileno)
			self.inventory_selection = j
			game.changed()

	def have_funds(self, amount):
		if amount <= self.funds:
			return True
		else:
			self.stop()
			skint_sound.play()
			alert("Out of funds!")
			return False
	
	def spend_funds(self, amount):
		if self.have_funds(amount):
			self.funds -= amount
			game.changed()
			return True
		else:
			return False
	
	def spend_funds_with_sound(self, amount):
		result = self.spend_funds(amount)
		if result:
			money_sound.play()
		return result
	
	def buy_random_pipe(self):
		if self.room_in_inventory():
			tile = random.choice(tiles.random_pipes)
			if self.spend_funds(tile.cost):
				money_sound.play()
				self.add_tile_to_inventory(tile)

	def timestep(self, dt):
		self.update_status()
		if self.running:
			if self.able_to_run() and self.spend_funds(RUNNING_COST):
				self.run()
			else:
				self.stop()
	
	def update_status(self):
		self.circuit_complete = self.trace_beam()
#		if not self.able_to_run():
#			self.stop()
	
	def trace_beam(self):
		start = (self.start_cell, SIDE_N)
		state = start
		self.beam_energy = 0
		self.vacuum_ok = True
		while 1:
			self.beam_energy += PIPE_ENERGY
			if not self.check_vacuum(state[0]):
				self.vacuum_ok = False
			state = self.step_beam(state)
			if not state:
				return False
			if state == start:
				return True
	
	def check_vacuum(self, cell):
		#print "World.check_vacuum:", cell ###
		tile = self.get_tile(cell)
		for dir in tile.exits():
			#print "...in direction", dir ###
			neighbour_cell = add(cell, direction_vectors[dir])
			#print "......neighbour cell =", neighbour_cell
			neighbour_tile = self.get_tile(neighbour_cell)
			#print "......neighbour tile =", neighbour_tile
			if not neighbour_tile:
				return False
			rid = opposite_side(dir)
			if not neighbour_tile.has_exit(rid):
				#print "......no exit in direction", rid ###
				return False
			#print "......has exit in direction", rid ###
		return True
	
	def step_beam(self, state):
		cell, dir = state
		cell = add(cell, direction_vectors[dir])
		tile = self.get_tile(cell)
		if not tile:
			return None
		dir = tile.exit_for_entrance(dir)
		if dir == SIDE_NONE:
			return None
		return (cell, dir)
	
	def run(self):
		if self.beam_energy >= self.energy_required:
			particle = particles.get(self.particles_found)
			if random.random() <= particle.discovery_chance:
				self.stop()
				success_sound.play()
				particle.announce_discovery()
				self.particles_found += 1
				game.changed()
				if particle.is_real:
					self.new_level()
				else:
					self.start()
	
	@property
	def beam_status(self):
		if self.circuit_complete:
			return "Complete"
		else:
			return "Incomplete"
	
	@property
	def vacuum_status(self):
		if self.vacuum_ok:
			return "Yes"
		else:
			return "No"
	
	@property
	def energy_required(self):
		return self.energy_required_for_level(self.level_number)
	
	def energy_required_for_level(self, n):
		if n == 0:
			return 0
		k = 0.4 + 0.1 * (n - 1) / 14
		return int(round(k * (n + 6) ** 2))
	
	def funds_required_for_level(self, n):
		e = self.energy_required_for_level(n) - self.energy_required_for_level(n-1)
		p = ceil(e / PIPE_ENERGY)
		c = ceil(2 * p / (n + 6))
		a = p * PIPE_COST + c * CORNER_COST + (p + c) * PLACEMENT_COST
		k = 1.10 - 0.05 * (n - 1) / 14
		m = ceil(a * k / 1000)
		return int(m) * 1000
	
	def dump_levels(self):
		mt = 0
		print "%5s %5s %12s" % ("Level", "Energy", "Funds")
		for n in xrange(1, 16):
			e = self.energy_required_for_level(n)
			m = self.funds_required_for_level(n)
			mt += m
			print "%5d %5d %12d" % (n, e, m)
		print "%5s %5s %12s" % ("", "", mt)
	
	def able_to_run(self):
		#print "World.able_to_run", self.circuit_complete ###
		#traceback.print_stack()
		return self.circuit_complete and self.vacuum_ok
	
	def is_startable(self):
		return self.able_to_run() and not self.running
	
	def is_running(self):
		return self.running
	
	def start(self):
		if self.able_to_run():
			self.running = True
			game.changed()
			#print "TODO: World.start: Start playing running sound" ###
			switch_on_sound.play()
			beam_sound.play(loops = -1)
	
	def stop(self):
		if self.running:
			self.running = False
			game.changed()
			#print "TODO: World.start: Stop playing running sound" ###
			switch_off_sound.play()
			beam_sound.fadeout(100)

#-----------------------------------------------------------------------------

def make_grid():
	z = TILE_NONE
	return [[z] * GRID_COLS for i in xrange(GRID_ROWS)]

def copy_grid(grid):
	return [row[:] for row in grid]

def grid_bounds_for_level(n):
	w = h = LEVEL1_GRID_SIZE + 2 * n
	if h > GRID_ROWS:
		w += (GRID_ROWS - h) // 2
		h = GRID_ROWS
	return Rect(0, (GRID_ROWS - h) // 2, w, h)

def announce_success():
	game_over_sound.play()
	result = ask("Excellent! You've won a Nobel Prize for finding all the supersymmetric partners.",
		["Modestly Decline", "Open Champagne"])
	if result == "Open Champagne":
		champagne_sound.play()
